/* 
 * ASCD : copyright (c) Aley Keprt 1998-2001
 *
 * sna.c - 48/128k .sna snap file handling
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <stdio.h>
#include "zlib.h"

#include "z80.h"
#include "mem.h"
#include "sna.h"


typedef unsigned char byte;
typedef unsigned short word;


extern byte *realMem[];

extern int dialog_running;


typedef struct {
  byte i,
    hl1[2],de1[2],bc1[2],af1[2],
    hl[2],de[2],bc[2],iy[2],ix[2],
    iff,r,
    af[2],sp[2],
    im,border;
} SNA_header;


int load_rom_image(char *image, int page); //main.c
void set_zx_palette(void);


//-----------------------------------------------------------------------------------


#define mw(reg) (r.reg=sna.reg[0]+sna.reg[1]*256)
#define mb(h,l,hl) (r.h=sna.hl[1],r.l=sna.hl[0])

static SNA_header sna;


static void initregs(int pc) {
//
//initialize registers
//
//pc: PC reg (-1 = on stack)
//
 Z80_Regs r;

 mw(ix);
 mw(iy);
 mw(sp);

 mb(a,f,af);
 mb(b,c,bc);
 mb(d,e,de);
 mb(h,l,hl);

 mb(a1,f1,af1);
 mb(b1,c1,bc1);
 mb(d1,e1,de1);
 mb(h1,l1,hl1);

 if(sna.im>=1) sna.im++; //we have im0123, .sna uses im0x12!!!
 r.im=sna.im;
 r.r=sna.r;
 r.i=sna.i;

 r.iff1=r.iff2=(sna.iff!=0);

 out_byte(BORDER_PORT,sna.border);

 if(pc==-1) {
   r.pc=read_word(r.sp);
   r.sp+=2;
 } else {
   r.pc=pc;
 }

 set_registers(&r);
}


int load_sna_gzf(gzFile *f) {
//
//internal function
//loads .sna snapshot - 48k or 128k version
//
//return: error flag
//
  byte page3[16384];
  int i;

  struct {
    word pc;
    byte page;
    byte trdos;
  } hdr2;

  if(gzread(f,&sna,27)!=27) return -1; //error
  if(gzread(f,realMem[5],16384)!=16384) return -1; //error
  if(gzread(f,realMem[2],16384)!=16384) return -1; //error
  if(gzread(f,page3,16384)!=16384) return -1; //error

  //128k header
  switch(gzread(f,&hdr2,4)) {

  //4 bytes read = ok
  case 4:
    break;

  //nothing read = 48k version
  case 0:
    //presune posledni stranku z bufferu
    memcpy(realMem[0],page3,16384);

    set_emulation_mode(EM_48k);
    initregs(-1);
    return 0;

  default:
    //error
    return -1;
  }

  //128k version
  set_emulation_mode(EM_128k);
  initregs(hdr2.pc);
  out_zxpage(hdr2.page);

  //nahraje ostatni ram stranky
  for(i=0;i<7;i++) {
    if(i!=2 && i!=5 && i!=(hdr2.page&7)) {
      if(gzread(f,realMem[i],16384)!=16384) return -1; //error
    }
  }

  //presune posledni stranku z bufferu
  memcpy(realMem[hdr2.page&7],page3,16384);

  return 0;
}


//-----------------------------------------------------------------------

#undef mw
#define mw(reg) (sna.reg[0]=r.reg%256,sna.reg[1]=r.reg/256)

#undef mb
#define mb(h,l,hl) (sna.hl[1]=r.h,sna.hl[0]=r.l)


int save_sna(const char *fname) {

  SNA_header sna;
  Z80_Regs r;
  byte dump[49152];
  FILE *f;

  get_registers(&r);

  //store registers
  mw(ix);
  mw(iy);
  r.sp-=2;
  mw(sp);

  mb(a,f,af);
  mb(b,c,bc);
  mb(d,e,de);
  mb(h,l,hl);

  mb(a1,f1,af1);
  mb(b1,c1,bc1);
  mb(d1,e1,de1);
  mb(h1,l1,hl1);

  sna.im=r.im;
  if(sna.im>=2) sna.im--; //we have im0123, .sna has im0x12!!!
  sna.r=r.r;
  sna.i=r.i;

  sna.iff=r.iff1+r.iff2*2;

  sna.border=border&7;

  write_word(r.sp,r.pc);

  memcpy(&dump[0],realMem[1],16384);
  memcpy(&dump[16384],realMem[2],16384);
  memcpy(&dump[32768],realMem[3],16384);

  //write .sna
  f=fopen(fname,"wb");
  fwrite(&sna,27,1,f);
  fwrite(dump,49152,1,f);
  fclose(f);

  return 0;
}


int load_sna(const char *fname) {
//
//load .sna snapshot - 48k or 128k version
//
//return: error flag
//
  gzFile *gzf;
  int r;

  gzf=gzopen(fname,"rb");
  if(!gzf) return 0;

  r=load_sna_gzf(gzf);
  gzclose(gzf);
  return r;
}

